// BinEditView.cpp : implementation of the CBinEditView class
//

#include "stdafx.h"
#include "BinEdit.h"

#include "BinEditDoc.h"
#include "BinEditView.h"
#include "MainFrm.h"
#include "ChangeDlg.h"
#include "FontDlg.h"

#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif

/////////////////////////////////////////////////////////////////////////////
// CBinEditView

IMPLEMENT_DYNCREATE(CBinEditView, CView)

BEGIN_MESSAGE_MAP(CBinEditView, CView)
	//{{AFX_MSG_MAP(CBinEditView)
	ON_WM_SIZE()
	ON_WM_CHAR()
	ON_WM_KEYDOWN()
	ON_WM_KEYUP()
	ON_WM_LBUTTONDOWN()
	ON_WM_VSCROLL()
	ON_UPDATE_COMMAND_UI(ID__BIN, OnUpdateBin)
	ON_UPDATE_COMMAND_UI(ID__DEC, OnUpdateDec)
	ON_UPDATE_COMMAND_UI(ID__HEX, OnUpdateHex)
	ON_UPDATE_COMMAND_UI(ID__OCT, OnUpdateOct)
	ON_COMMAND(ID__BIN, OnBin)
	ON_COMMAND(ID__DEC, OnDec)
	ON_COMMAND(ID__HEX, OnHex)
	ON_COMMAND(ID__OCT, OnOct)
	ON_UPDATE_COMMAND_UI(ID_GOTOADDR, OnUpdateGotoaddr)
	ON_COMMAND(ID_GOTOADDR, OnGotoaddr)
	ON_UPDATE_COMMAND_UI(ID_CHANGEVAL, OnUpdateChangeval)
	ON_COMMAND(ID_CHANGEVAL, OnChangeval)
	ON_WM_LBUTTONDBLCLK()
	ON_COMMAND(ID_FONTSIZE, OnFontsize)
	ON_UPDATE_COMMAND_UI(ID___DOS, OnUpdateDos)
	ON_COMMAND(ID___DOS, OnDos)
	ON_UPDATE_COMMAND_UI(ID___WINDOWS, OnUpdateWindows)
	ON_COMMAND(ID___WINDOWS, OnWindows)
	//}}AFX_MSG_MAP
END_MESSAGE_MAP()

/////////////////////////////////////////////////////////////////////////////
// CBinEditView construction/destruction

CBinEditView::CBinEditView()
{
	pFont = new CFont;
	m_Control = false;
	m_nCharHeight = 16;
	m_nSystem = 3;
	m_nNumWidth = 2;
	textbuf = NULL;
	data = NULL;
	MakeFont(m_nCharHeight);
	m_AddrDlg.m_Addr = 0;
	m_AddrDlg.m_Radix = 0;
	codetable = new unsigned char[256];
	m_nCodePage = 1;
	BuildCodeTable(m_nCodePage);
	m_SymDlg.codetable = codetable;
}

CBinEditView::~CBinEditView()
{
	delete pFont;
	delete codetable;
}

BOOL CBinEditView::PreCreateWindow(CREATESTRUCT& cs)
{
	cs.style |= WS_VSCROLL;
	WNDCLASSEX wClass;
	GetClassInfoEx(cs.hInstance, cs.lpszClass, &wClass);
	wClass.style = 0;
	wClass.hbrBackground = 0;
	return CView::PreCreateWindow(cs);
}

/////////////////////////////////////////////////////////////////////////////
// CBinEditView drawing

void CBinEditView::OnDraw(CDC* pDC)
{
	CDC dc;
	CBitmap bm, *old_bm;
	CRect r;
	CPen p1, p2, p3;
	CBrush br;

	if (!textbuf) return;

	GetClientRect(&r);
	dc.CreateCompatibleDC(pDC);
	bm.CreateCompatibleBitmap(pDC, r.Width(), r.Height());
	p2.CreatePen(PS_SOLID, 1, GetSysColor(COLOR_WINDOW));
	br.CreateSysColorBrush(COLOR_WINDOW);
	old_bm = dc.SelectObject(&bm);
	dc.SelectObject(&p2);
	dc.SelectObject(&br);
	dc.Rectangle(&r);
	dc.SelectObject(pFont);
	dc.SetBkMode(TRANSPARENT);
	for (int i = 0, a = 0, b = 0; i < m_nRows; i++, a += m_nCharHeight, b += m_nBufWidth)
		dc.TextOut(0, a, CString(textbuf + b));
	dc.SelectStockObject(SYSTEM_FONT);

	int x = (m_nBufWidth - m_nCols - 3) * m_nCharWidth;
	
	p1.CreatePen(PS_SOLID, 1, GetSysColor(COLOR_BTNFACE));
	p3.CreatePen(PS_SOLID, 1, GetSysColor(COLOR_BTNSHADOW));

	dc.SelectObject(p1);
	dc.MoveTo(CPoint(x, 0));
	dc.LineTo(CPoint(x, r.Height()));
	dc.SelectObject(p2);
	dc.MoveTo(CPoint(x + 1, 0));
	dc.LineTo(CPoint(x + 1, r.Height()));
	dc.SelectObject(p1);
	dc.MoveTo(CPoint(x + 2, 0));
	dc.LineTo(CPoint(x + 2, r.Height()));
	dc.MoveTo(CPoint(x + 3, 0));
	dc.LineTo(CPoint(x + 3, r.Height()));
	dc.SelectObject(p3);
	dc.MoveTo(CPoint(x + 4, 0));
	dc.LineTo(CPoint(x + 4, r.Height()));
	dc.SelectStockObject(BLACK_PEN);
	dc.MoveTo(CPoint(x + 5, 0));
	dc.LineTo(CPoint(x + 5, r.Height()));

	pDC->BitBlt(0, 0, r.Width(), r.Height(), &dc, 0, 0, SRCCOPY);
	dc.SelectStockObject(WHITE_BRUSH);
	dc.SelectObject(old_bm);
}

/////////////////////////////////////////////////////////////////////////////
// CBinEditView diagnostics

#ifdef _DEBUG
void CBinEditView::AssertValid() const
{
	CView::AssertValid();
}

void CBinEditView::Dump(CDumpContext& dc) const
{
	CView::Dump(dc);
}

CBinEditDoc* CBinEditView::GetDocument() // non-debug version is inline
{
	ASSERT(m_pDocument->IsKindOf(RUNTIME_CLASS(CBinEditDoc)));
	return (CBinEditDoc*)m_pDocument;
}
#endif //_DEBUG

void CBinEditView::NumberToText(unsigned char n, unsigned char *text)
{
	int i;

	switch (m_nSystem)
	{
		case 0: 
			for (i = 0x80; i; i >>= 1, text++) 
				*text = n & i ? '1' : '0'; 
			break;
		case 1: 
			for (i = 0; i < 3; i++, n >>= 3, text--) 
				*(text + 2) = (n & 7) + '0'; 
			break;
		case 2: 
			text[0] = n / 100 + '0'; 
			text[1] = (n % 100) / 10 + '0'; 
			text[2] = (n % 10) + '0'; 
			break;
		case 3:
			text[0] = hex_table[n >> 4];
			text[1] = hex_table[n & 0x0F];
	}
}

void CBinEditView::MakeTextBuf()
{
	CBinEditDoc *pDoc = GetDocument();
	int i, j, m, n;
	int buf_size = m_nBufWidth * m_nRows;
	int k = m_nBufWidth - m_nCols - 1, p = m_nPagePos;
	unsigned char *t, *t2, *d;

	if (data) delete data;
	data = new unsigned char[m_nRows * m_nCols];
	pDoc->GetData(m_nPagePos, data, m_nPageSize);
	d = data;

	if (textbuf) delete textbuf;
	textbuf = new unsigned char[buf_size];
	t = t2 = textbuf;

	memset(t, ' ', buf_size);
	t2 = t + m_nBufWidth - 1;
	for (i = 0; i < m_nRows; i++, t2 += m_nBufWidth) *t2 = 0x00;
	n = min(m_nFileSize - m_nPagePos, m_nPageSize);

	for (i = 0; i < n; i++)
	{
		j = i % m_nCols;
		if (!j)
		{
			if (i)
			{
				p += m_nCols;
				d += m_nCols;
			}
			t2 = t;
			for (m = 28; m >= 0; m -= 4, t2++)
				*t2 = hex_table[(p >> m) & 0x0F];
			*t2 = ':';
			t2 += 2;
		}
		NumberToText(d[j], t2);		
		t[k + j] = codetable[d[j]];
		t2 += m_nNumWidth + 1;
		if (j == m_nCols - 1) t += m_nBufWidth;
	}
}

void CBinEditView::CalcVars(int width, int height)
{
	m_nRows = height / m_nCharHeight;
	m_nCols = (width / m_nCharWidth - 12) / (m_nNumWidth + 2);
	if (m_nRows < 2) m_nRows = 2;
	if (m_nCols < 2) m_nCols = 2;
	m_nPageSize = m_nRows * m_nCols;
	m_nBufWidth = (m_nNumWidth + 2) * m_nCols + 13;
}

void CBinEditView::CheckVars()
{
	if (m_nFileSize - m_nPagePos < m_nPageSize)
		m_nPagePos = m_nFileSize - m_nPageSize;
	if (m_nPagePos < 0) 
		m_nPagePos = 0;
	if (m_nCursorPos >= m_nFileSize)
		m_nCursorPos = m_nFileSize - 1;
	if (m_nCursorPos >= m_nPageSize)
		m_nCursorPos = m_nPageSize - 1;
	if (m_nCursorPos < 0)
		m_nCursorPos = 0;
}

void CBinEditView::MakeCaret()
{
	int w;
	if (!m_nActivePane)
		w = m_nNumWidth * m_nCharWidth;
	else
		w = m_nCharWidth;
	CreateSolidCaret(w, m_nCharHeight);
}

void CBinEditView::SetCaret()
{
	int x, y;

	if (!m_nFileSize) return;
	x = m_nCursorPos % m_nCols;
	y = m_nCursorPos / m_nCols;
	if (!m_nActivePane)
		x = (x * (m_nNumWidth + 1) + 10) * m_nCharWidth;
	else
		x = (m_nBufWidth - m_nCols - 1 + x) * m_nCharWidth;
	y = y * m_nCharHeight;

	SetCaretPos(CPoint(x, y));
	ShowCaret();
}

void CBinEditView::UpdateStatus()
{
	CStatusBar *s = &((CMainFrame *)AfxGetMainWnd())->m_wndStatusBar;
	int pos = m_nPagePos + m_nCursorPos;
	char p[12], t[10];
	unsigned char c = data[m_nCursorPos];
	itoa(c, t, 2);
	sprintf(p, "%08s", t);
	s->SetPaneText(1, p);
	itoa(c, t, 8);
	sprintf(p, "%03s", t);
	s->SetPaneText(2, p);
	sprintf(p, "%03d", c);
	s->SetPaneText(3, p);
	sprintf(p, "%02X", c);
	s->SetPaneText(4, p);
	sprintf(p, "%08X", pos);
	s->SetPaneText(5, p);
	sprintf(p, "%010ld", pos);
	s->SetPaneText(6, p);
}

void CBinEditView::SetSB()
{
	SCROLLINFO si;

	if (m_nFileSize < m_nPageSize)
	{
		EnableScrollBar(SB_VERT, ESB_DISABLE_BOTH);
		return;
	}
	else
		EnableScrollBar(SB_VERT, ESB_ENABLE_BOTH);

	si.cbSize = sizeof(si);
	si.fMask = SIF_ALL;
	si.nMax = 32767;
	si.nMin = 0;
	si.nPage = m_nPageSize * 32767 / m_nFileSize;
	if (!si.nPage) si.nPage = 1;
	si.nPos = 0;
	si.nTrackPos = 0;
	SetScrollInfo(SB_VERT, &si);
}

void CBinEditView::SetSBPos()
{
	int pos = m_nPagePos * 32767 / m_nFileSize;
	SetScrollPos(SB_VERT, pos);
}

void CBinEditView::MakeFont(int height)
{
	pFont->Detach();
	pFont->CreateFont(height, 0, 0, 0, FW_DONTCARE, 0, 0, 0, 
		DEFAULT_CHARSET, OUT_DEFAULT_PRECIS, CLIP_DEFAULT_PRECIS,
		DEFAULT_QUALITY, DEFAULT_PITCH, "Courier New Cyr");
	CDC dc;
	TEXTMETRIC tm;
	dc.CreateDC("DISPLAY", NULL, NULL, NULL);
	dc.SelectObject(pFont);
	dc.GetTextMetrics(&tm);
	m_nCharWidth = tm.tmAveCharWidth;
	dc.SelectStockObject(SYSTEM_FONT);
}

void CBinEditView::SetSystem(int sys)
{
	if (m_nSystem == sys) return;

	switch (sys)
	{
		case 0: m_nNumWidth = 8; break;
		case 1: m_nNumWidth = 3; break;
		case 2: m_nNumWidth = 3; break;
		case 3: m_nNumWidth = 2;
	}

	CRect r;
	GetClientRect(&r);
	m_nSystem = sys;
	CalcVars(r.Width(), r.Height());
	CheckVars();
	MakeTextBuf();
	MakeCaret();
	SetCaret();
	SetSB();
	SetSBPos();
	InvalidateRect(NULL, FALSE);
}

void CBinEditView::BuildCodeTable(int type)
{
	int i = 1;
	char s[] = "___";

	codetable[0] = ' ';
	for (; i < 256; i++)
		codetable[i] = i;

	if (!type)
	{
		for (i = 224; i < 240; i++)
			codetable[i] += 16;
		for (i = 128; i < 176; i++)
			codetable[i] += 64;
		for (i = 192; i < 224; i++)
			codetable[i] -= 64;
		memcpy(codetable + 240, s, 16);
	}
}

void CBinEditView::ChangeCodePage(int new_code)
{
	m_nCodePage = new_code;
	BuildCodeTable(new_code);
	MakeTextBuf();
	InvalidateRect(NULL, FALSE);
}

/////////////////////////////////////////////////////////////////////////////
// CBinEditView message handlers

void CBinEditView::OnSize(UINT nType, int cx, int cy) 
{
	CView::OnSize(nType, cx, cy);

	if (!cx || !cy) return;

	CalcVars(cx, cy);
	CheckVars();
	MakeTextBuf();
	SetCaret();
	InvalidateRect(NULL, FALSE);
}


void CBinEditView::OnActivateView(BOOL bActivate, CView* pActivateView, CView* pDeactiveView) 
{
	if (bActivate)
	{
		MakeCaret();
		SetCaret();
	}
	else
		DestroyCaret();
	
	CView::OnActivateView(bActivate, pActivateView, pDeactiveView);
}

void CBinEditView::OnChar(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	CView::OnChar(nChar, nRepCnt, nFlags);
}

void CBinEditView::OnKeyDown(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	int old_cur = m_nCursorPos;
	int old_page = m_nPagePos;

	switch (nChar)
	{
		case VK_TAB: 
			m_nActivePane ^= 1; 
			MakeCaret();
			SetCaret(); 
			return;
		case VK_CONTROL: m_Control = true; break;
		case VK_HOME: 
			if (m_Control)
			{
				m_nCursorPos = 0;
				m_nPagePos = 0;
			}
			else
				m_nCursorPos -= m_nCursorPos % m_nCols; 
			break;
		case VK_END: 
			if (m_Control)
			{
				m_nCursorPos = m_nPageSize - 1;
				m_nPagePos = m_nFileSize - m_nPageSize;
			}
			else
				m_nCursorPos = m_nCursorPos - (m_nCursorPos % m_nCols) + m_nCols - 1; 
			break;
		case VK_LEFT:
			if (m_Control) m_nPagePos--;
			else m_nCursorPos--; 
			break;
		case VK_RIGHT:
			if (m_Control) m_nPagePos++;
			else m_nCursorPos++; 
			break;
		case VK_UP: m_nCursorPos -= m_nCols; break;
		case VK_DOWN: m_nCursorPos += m_nCols; break;
		case VK_NEXT: m_nPagePos += m_nPageSize; break;
		case VK_PRIOR:m_nPagePos -= m_nPageSize; 
	}

	if (m_nCursorPos >= m_nPageSize) 
	{
		m_nPagePos += m_nCols;
		m_nCursorPos -= m_nCols;
	}
	if (m_nCursorPos < 0) 
	{
		m_nPagePos -= m_nCols;
		m_nCursorPos += m_nCols;
	}
	if (m_nPagePos < 0) m_nCursorPos = 0;
	if (m_nFileSize - m_nPagePos < m_nPageSize && m_nFileSize > m_nPageSize ||
		m_nFileSize <= m_nPageSize && nChar == VK_NEXT) 
		m_nCursorPos = m_nPageSize - 1;

	CheckVars();

	if (old_cur != m_nCursorPos) SetCaret();
	if (old_page != m_nPagePos) 
	{
		SetSBPos();
		MakeTextBuf();
		InvalidateRect(NULL, FALSE);
	}
	if (old_cur != m_nCursorPos || old_page != m_nPagePos) 
		UpdateStatus();
	
	CView::OnKeyDown(nChar, nRepCnt, nFlags);
}

void CBinEditView::OnKeyUp(UINT nChar, UINT nRepCnt, UINT nFlags) 
{
	if (nChar == VK_CONTROL)
		m_Control = false;
	
	CView::OnKeyUp(nChar, nRepCnt, nFlags);
}

void CBinEditView::OnUpdate(CView* pSender, LPARAM lHint, CObject* pHint) 
{
	CBinEditDoc *pDoc = GetDocument();
	char size[20];
	m_nFileSize = pDoc->GetFileSize();
	CStatusBar *s = &((CMainFrame *)AfxGetMainWnd())->m_wndStatusBar;
	itoa(m_nFileSize, size, 10);
	s->SetPaneText(7, size);
	MakeTextBuf();
	InvalidateRect(NULL, FALSE);
	MakeCaret();
	SetCaret();
	SetSB();
	m_nActivePane = 0;
	m_nPagePos = 0;
	m_nCursorPos = 0;
	if (m_nFileSize) 
		UpdateStatus();
	else
		for (int i = 1; i < 7; i++)
			s->SetPaneText(i, "");
}

int CBinEditView::SetMouseSel(CPoint point)
{
	if (point.x < m_nCharWidth * 10) return 0;
	if (point.x > m_nCharWidth * (m_nBufWidth - 1)) return 0;
	if (point.y > m_nCharHeight * m_nRows) return 0;

	int row = point.y / m_nCharHeight, col, pos;
	int old_pane = m_nActivePane;

	if (point.x < m_nCharWidth * (m_nBufWidth - m_nCols - 3))
	{
		col = (point.x / m_nCharWidth - 10) / (m_nNumWidth + 1);
		m_nActivePane = 0;
	}
	else 
	{
		col = m_nCharWidth * (m_nBufWidth - m_nCols - 1);
		if (point.x >= col)
		{
			col = (point.x - col) / m_nCharWidth;
			m_nActivePane = 1;
		}
		else 
			return 0;
	}
	
	pos = row * m_nCols + col;
	if (pos >= m_nFileSize - m_nPagePos) return 0;
	m_nCursorPos = pos;
	CheckVars();
	if (old_pane != m_nActivePane) MakeCaret();
	SetCaret();
	UpdateStatus();

	return 1;
}

void CBinEditView::OnLButtonDown(UINT nFlags, CPoint point) 
{
	SetMouseSel(point);
	CView::OnLButtonDown(nFlags, point);
}

void CBinEditView::OnVScroll(UINT nSBCode, UINT nPos, CScrollBar* pScrollBar) 
{
	int old_page = m_nPagePos;

	switch (nSBCode)
	{
		case SB_LINEDOWN: m_nPagePos += m_nCols; break;
		case SB_LINEUP: m_nPagePos -= m_nCols; break;
		case SB_PAGEDOWN: m_nPagePos += m_nPageSize; break;
		case SB_PAGEUP: m_nPagePos -= m_nPageSize; break;
		case SB_THUMBTRACK: 
			m_nPagePos = int((double)nPos / 32767 * m_nFileSize);
			m_nPagePos = m_nPagePos / m_nCols * m_nCols;
	}
	
	CheckVars();
	SetSBPos();

	if (old_page != m_nPagePos) 
	{
		MakeTextBuf();
		InvalidateRect(NULL, FALSE);
	}
	
	CView::OnVScroll(nSBCode, nPos, pScrollBar);
}

void CBinEditView::OnUpdateBin(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(!m_nSystem);
}

void CBinEditView::OnUpdateDec(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(m_nSystem == 2);
}

void CBinEditView::OnUpdateHex(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(m_nSystem == 3);
}

void CBinEditView::OnUpdateOct(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(m_nSystem == 1);
}

void CBinEditView::OnBin() 
{
	SetSystem(0);	
}

void CBinEditView::OnDec() 
{
	SetSystem(2);	
}

void CBinEditView::OnHex() 
{
	SetSystem(3);	
}

void CBinEditView::OnOct() 
{
	SetSystem(1);	
}

void CBinEditView::OnUpdateGotoaddr(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(m_nFileSize);
}

void CBinEditView::OnGotoaddr() 
{
	if (m_AddrDlg.DoModal() != IDOK) return;
	m_nPagePos = m_AddrDlg.m_Addr;
	m_nCursorPos = 0;
	if (m_nFileSize - m_nPagePos < m_nPageSize)
	{
		m_nPagePos = m_nFileSize - m_nPageSize;
		m_nCursorPos += m_AddrDlg.m_Addr - m_nPagePos;
		if (m_nCursorPos >= m_nPageSize)
			MessageBox("    ");
	}
	CheckVars();
	SetCaret();
	SetSBPos();
	MakeTextBuf();
	UpdateStatus();
	InvalidateRect(NULL, FALSE);
}

void CBinEditView::OnUpdateChangeval(CCmdUI* pCmdUI) 
{
	pCmdUI->Enable(m_nFileSize);
}

void CBinEditView::OnChangeval() 
{
	CBinEditDoc *pDoc = GetDocument();
	if (!m_nActivePane)
	{
		CChangeDlg d;
		d.m_Radix = m_nSystem;
		d.m_Val = data[m_nCursorPos];
		if (d.DoModal() != IDOK) return;
		pDoc->WriteByte(m_nPagePos + m_nCursorPos, d.m_Val);
	}
	else
	{
		m_SymDlg.m_Sym = data[m_nCursorPos];
		if (m_SymDlg.DoModal() != IDOK) return;
		pDoc->WriteByte(m_nPagePos + m_nCursorPos, m_SymDlg.m_Sym);
	}
	MakeTextBuf();
	InvalidateRect(NULL, FALSE);
}

void CBinEditView::OnLButtonDblClk(UINT nFlags, CPoint point) 
{
	if (SetMouseSel(point))
		OnChangeval();
	CView::OnLButtonDblClk(nFlags, point);
}

void CBinEditView::OnFontsize() 
{
	CFontDlg d;
	char s[10];
	itoa(m_nCharHeight, s, 10);
	d.m_FontSize = CString(s);
	if (d.DoModal() != IDOK) return;
	m_nCharHeight = atoi(d.m_FontSize);
	MakeFont(m_nCharHeight);
	CRect r;
	GetClientRect(&r);
	CalcVars(r.Width(), r.Height());
	CheckVars();
	MakeTextBuf();
	MakeCaret();
	SetCaret();
	SetSB();
	SetSBPos();
	InvalidateRect(NULL, FALSE);
}

void CBinEditView::OnUpdateDos(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(!m_nCodePage);
}

void CBinEditView::OnDos() 
{
	ChangeCodePage(0);
}

void CBinEditView::OnUpdateWindows(CCmdUI* pCmdUI) 
{
	pCmdUI->SetCheck(m_nCodePage);
}

void CBinEditView::OnWindows() 
{
	ChangeCodePage(1);
}
